from datetime import timedelta
import os

from ._backend import Backend


class SlurmBackend(Backend):
    def __init__(self):
        super().__init__()
        self.name = 'slurm'
        self.task_id_var = r'$SLURM_ARRAY_TASK_ID'

    def get_cancel_cmds(self, cancellations):
        cmds = []
        for cancellation in cancellations:
            jobid, tasklist = cancellation
            cmd = "scancel"
            if tasklist is None:
                cmd += " {}".format(jobid)
            else:
                for task_id in tasklist:
                    cmd += " {}_{}".format(jobid, task_id)
            cmds.append(cmd)
        return cmds

    def get_job_list(self, args):
        # Call the appropriate sbatch command. The default behavior is to use
        # Slurm's job array feature, which starts a batch job with multiple tasks
        # and passes a different taskid to each one. If ntasks is zero, only a
        # single job is submitted with no subtasks.
        base_cmd = 'sbatch '
        base_cmd += '-J {} '.format(args.jobname)  # set name of job
        # Slurm runs scripts in current working directory by default

        # Duration
        base_cmd += '-t {} '.format(args.duration)
        duration = self.get_time_delta(args.duration)

        # Number of CPU/GPU resources
        base_cmd += '-n {} '.format(args.cpus)
        if args.debug and duration > timedelta(hours=2):
            raise RuntimeError('{}: Duration cannot exceed 2 hours while in debug/test mode.'.format(self.name))
        if args.gpus > 0:
            partition = 'gpu-debug' if args.debug else 'gpu'
            base_cmd += '-p {} --gres=gpu:{} '.format(partition, args.gpus)
        else:
            partition = 'debug' if duration <= timedelta(hours=1) else 'batch'
            base_cmd += '-p {} '.format(partition)

        # Memory requirements
        base_cmd += '--mem={}G '.format(args.mem)

        # Logging
        log_dir = self.get_log_dir()
        # Format is jobname_jobid_taskid.*
        base_cmd += '-o {} '.format(os.path.join(log_dir, '%x_%A_%a.o'))  # save stdout to file
        base_cmd += '-e {} '.format(os.path.join(log_dir, '%x_%A_%a.e'))  # save stderr to file

        # The --parsable flag causes sbatch to print the jobid to stdout. We read the
        # jobid with subprocess.check_output()
        base_cmd += '--parsable '

        base_cmd += "--array={}".format(args.tasklist)
        if args.maxtasks > 0:
            # set maximum number of running tasks
            base_cmd += '%{} '.format(args.maxtasks)
        else:
            base_cmd += ' '

        # Prevent Slurm from running this new job until the specified job ID is finished.
        if args.hold_jid is not None:
            base_cmd += "--depend=afterany:{} ".format(args.hold_jid)

        wrapper_script = self.wrap_tasks(args.jobfile)
        wrapper_file = self.save_wrapper_script(wrapper_script, args.jobname)
        base_cmd += "{}".format(wrapper_file)
        return [base_cmd]



from collections import namedtuple
import subprocess
import sys

from tabulate import tabulate

from .utils import load_jobindex, expand_ids, load_jobfile

JobListing = namedtuple('JobListing', ['job_id', 'task_id', 'jobname', 'command', 'tag'])

def get_job_listings(args):
    def in_joblist(job_id):
        return True if args.jobid is None else job_id == args.jobid
    def in_tasklist(task_id):
        return True if args.tasklist is None else task_id in expand_ids(args.tasklist)

    job_list = []
    try:
        index = load_jobindex()
    except (IOError):
        return job_list
    for jobid in index.keys():
        if not in_joblist(jobid):
            continue
        jobname, jobfile = index[jobid]
        commands, tags = load_jobfile(jobfile)
        for task_id in sorted(commands.keys()):
            if not in_tasklist(task_id):
                continue
            listing = JobListing(
                jobid,
                task_id,
                jobname,
                repr(commands[task_id]),
                tags[task_id]
            )
            job_list.append(listing)
    return job_list

def list_commands(args, quiet=False):
    """Parse the jobs/tasks to cancel and send the appropriate commands to the cluster"""
    job_list = get_job_listings(args)
    if not quiet:
        print(tabulate(job_list, headers=JobListing._fields))
    



from contextlib import ExitStack
import subprocess
import sys

from .utils import load_jobfile

def run_command_by_id(commands, task_id, stdout=None, stderr=None, quiet=False):
    cmd = commands[task_id]
    if not quiet:
        print('Launching worker:', cmd)
    with ExitStack() as stack:
        stdout = stack.enter_context(open(stdout, 'wb')) if stdout is not None else None
        stderr = stack.enter_context(open(stderr, 'wb')) if stderr is not None else None
        try:
            subprocess.call(cmd, shell=True, stdout=stdout, stderr=stderr)
        except:
            raise
        else:
            if not quiet:
                print('Worker finished:', cmd)

if __name__ == '__main__':
    assert len(sys.argv) == 3, 'Usage: python -m worker path/to/commands.json task_id'
    
    commands_file = sys.argv[1]
    task_id = int(sys.argv[2])

    commands = load_jobfile(commands_file)[0]

    run_command_by_id(commands, task_id)



from datetime import datetime, timedelta
import os
import subprocess
import sys

from ..config import get_active_config
from ..constants import default_logs_folder
from ..utils import update_jobindex, insert_second_to_last

class Backend:
    def __init__(self):
        self.name = 'generic_backend'
        self.header = '#!/bin/bash\n'
        self.body = '\npython -m onager.worker {} {} \n'
        self.footer = ''

        self.task_id_var = r'$TASK_ID'

    def wrap_tasks(self, tasks_file, stdout=None, stderr=None):
        config = get_active_config()
        header = self.header + config[self.name]['header']
        body = self.body.format(tasks_file, self.task_id_var)
        footer = config[self.name]['footer'] + self.footer
        wrapper_script = header + body + footer
        return wrapper_script

    def save_wrapper_script(self, wrapper_script, jobname):
        scripts_dir = ".onager/scripts/{}/".format(jobname)
        os.makedirs(scripts_dir, exist_ok=True)
        jobfile = os.path.join(scripts_dir, 'wrapper.sh')
        with open(jobfile, 'w') as file:
            file.write(wrapper_script)
        return jobfile

    def get_job_list(self, args):
        raise NotImplementedError

    def get_cancel_cmds(self, cancellations):
        raise NotImplementedError

    def get_time_delta(self, time_str):
        if '-' in time_str:
            days, hours_minutes_seconds = time_str.split('-')
        else:
            days, hours_minutes_seconds = 0, time_str
        t = datetime.strptime(hours_minutes_seconds, "%H:%M:%S")
        partial_days = timedelta(hours=t.hour, minutes=t.minute, seconds=t.second)
        return timedelta(days=int(days)) + partial_days

    def get_log_dir(self):
        log_dir = os.path.join(default_logs_folder, self.name)
        os.makedirs(log_dir, exist_ok=True)
        return log_dir

    def generate_tasklist(self, commands):
        ids = sorted(commands.keys())
        tasklist = ','.join(map(str, ids))
        return tasklist

    def launch(self, jobs, args, other_args):
        jobids = []

        if len(other_args) != 0:
            additional_args = ' '.join(other_args)
            jobs = [insert_second_to_last(job, additional_args) for job in jobs]

        for job in jobs:
            if not args.quiet:
                print(job)
            if not args.dry_run:
                try:
                    byte_str = subprocess.check_output(job, shell=True)
                    jobid = byte_str.decode('utf-8').split('.')[0]
                    jobids.append(jobid)
                except (subprocess.CalledProcessError, ValueError) as err:
                    print(err)
                    sys.exit()
        job_entries = [(jobid, args.jobname, args.jobfile) for jobid in jobids]
        update_jobindex(job_entries, append=True)



import math
from multiprocessing import Pool, cpu_count
import os
import socket

from ._backend import Backend
from ..worker import run_command_by_id
from ..utils import expand_ids, load_jobfile, update_jobindex, get_next_index_jobid

class LocalBackend(Backend):
    def __init__(self):
        super().__init__()
        hostname = socket.gethostname().replace('.local', '')
        self.name = hostname

    def get_job_list(self, args):
        return load_jobfile(args.jobfile)[0]

    def get_next_jobid(self):
        return 0

    def launch(self, jobs, args, other_args):
        self.commands = jobs
        if len(other_args) != 0:
            raise RuntimeError("{}: Cannot pass in additional args {}".format(
                self.name, ' '.join(other_args)))
        self.quiet = args.quiet
        log_name = '{}_{}'.format(args.jobname, self.get_next_jobid())
        self.log_path = os.path.join(self.get_log_dir(), log_name)
        os.makedirs(os.path.dirname(self.log_path), exist_ok=True)
        task_ids = expand_ids(args.tasklist)

        job_entries = [(get_next_index_jobid(), args.jobname, args.jobfile)]
        update_jobindex(job_entries, append=True)

        if args.maxtasks > 0:
            n_workers = args.maxtasks
        else:
            max_parallel = math.floor(cpu_count() / args.cpus)
            workers_available = max(1, max_parallel)
            workers_needed = len(task_ids)
            n_workers = min(workers_needed, workers_available)
        if not self.quiet:
            print('Starting multiprocessing pool with {} workers'.format(n_workers))
        pool = Pool(n_workers, maxtasksperchild=1)# Each new tasks gets a fresh worker
        pool.map(self.process_one_job, task_ids)
        pool.close()
        pool.join()

    def process_one_job(self, task_id):
        run_command_by_id(
            self.commands,
            task_id,
            stdout=self.log_path + '_{}.o'.format(task_id),
            stderr=self.log_path + '_{}.e'.format(task_id),
            quiet=self.quiet,
        )



from . import local, gridengine, slurm

def prepare_backend(args):
    if args.backend == 'local':
        backend = local.LocalBackend()
    elif args.backend == 'gridengine':
        backend = gridengine.GridEngineBackend()
    elif args.backend == 'slurm':
        backend = slurm.SlurmBackend()
    else:
        raise NotImplementedError('Invalid backend')
    return backend


from datetime import timedelta
import os

from ._backend import Backend
from ..utils import condense_ids

class GridEngineBackend(Backend):
    def __init__(self):
        super().__init__()
        self.name = 'gridengine'
        self.task_id_var = r'$SGE_TASK_ID'

    def generate_tasklist(self, commands):
        ids = sorted(commands.keys())
        tasklist = condense_ids(ids)
        return tasklist
    
    def get_cancel_cmds(self, cancellations):
        cmds = []
        for cancellation in cancellations:
            jobid, tasklist = cancellation
            tasklist = condense_ids(tasklist)
            cmd = "qdel {} ".format(jobid)
            if tasklist is not None:
                for taskblock in tasklist.split(','):
                    cmds.append(cmd + "-t {}".format(taskblock))
        return cmds

    def get_job_list(self, args):
        # Call the appropriate qsub command. The default behavior is to use
        # GridEngine's range feature, which starts a batch job with multiple tasks
        # and passes a different taskid to each one. If ntasks is zero, only a
        # single job is submitted with no subtasks.
        base_cmd = 'qsub '
        base_cmd += '-N {} '.format(args.jobname)  # set name of job
        base_cmd += '-cwd '  # run script in current working directory

        # Duration and number of CPU/GPU resources
        #
        # Note that the Brown CS grid grants all GPU jobs infinite duration
        #  -l test   (10 min, high priority, limited to one slot per machine)
        #  -l short  (1 hour)
        #  -l long   (1 day)
        #  -l vlong  (infinite duration)
        #  -l gpus=# (infinite duration, on a GPU machine)
        duration = self.get_time_delta(args.duration)
        if args.debug and duration > timedelta(minutes=10):
            raise RuntimeError('{}: Duration cannot exceed 10 minutes while in debug/test mode.'.format(self.name))
        if args.gpus > 0:
            queue = 'gpu'
            base_cmd += '-l gpus={} '.format(args.gpus)  # Request GPUs
        else:
            if args.debug:
                queue = 'test'
            elif duration > timedelta(days=1):
                queue = 'vlong'
            elif duration > timedelta(hours=1):
                queue = 'long'
            else:
                queue = 'short'
            base_cmd += '-l {} '.format(queue)
        if args.cpus > 1:
            if args.debug:
                raise RuntimeError('{}: Cannot use more than one cpu in debug/test mode. Try submitting a short-duration job instead.'.format(self.name))
            base_cmd += '-pe smp {} '.format(args.cpus)  # Request multiple CPUs

        # Memory requirements
        if args.mem > 1:
            base_cmd += '-l vf={}G '.format(args.mem)  # Reserve extra memory

        # if args.host is not None:
        #     base_cmd += '-q {}.q@{}.cs.brown.edu '.format(args.duration, args.host)

        # Logging
        log_dir = self.get_log_dir()
        # Format is jobname_jobid_taskid.*
        base_cmd += '-o {} '.format(os.path.join(
            log_dir, r'\$JOB_NAME_\$JOB_ID_\$TASK_ID.o'))  # save stdout to file
        base_cmd += '-e {} '.format(os.path.join(
            log_dir, r'\$JOB_NAME_\$JOB_ID_\$TASK_ID.e'))  # save stderr to file

        # The -terse flag causes qsub to print the jobid to stdout. We read the
        # jobid with subprocess.check_output()
        base_cmd += '-terse '

        # Prevent GridEngine from running this new job until the specified job ID is finished.
        if args.hold_jid is not None:
            base_cmd += "-hold_jid {} ".format(args.hold_jid)

        if args.maxtasks > 0:
            # set maximum number of running tasks per block
            base_cmd += "-tc {} ".format(args.maxtasks)

        wrapper_script = self.wrap_tasks(args.jobfile)
        wrapper_file = self.save_wrapper_script(wrapper_script, args.jobname)

        # Split tasklist into blocks that GridEngine can understand
        task_blocks = args.tasklist.split(',')
        return [
            base_cmd + "-t {} {}".format(task_block, wrapper_file) for task_block in task_blocks
        ]



import re
import os

from .backends import prepare_backend
from .constants import defaultjobfile
from .utils import load_jobfile, save_jobfile

def launch(args, other_args):
    if not re.match(r'^(\w|\.|-)+$', args.jobname):
        # We want to create a script file, so make sure the filename is legit
        raise ValueError("Invalid job name: {}".format(args.jobname))

    if args.jobfile == defaultjobfile:
        args.jobfile = args.jobfile.format(jobname=args.jobname)
    os.makedirs(os.path.dirname(args.jobfile), exist_ok=True)
    if args.command is not None:
        save_jobfile({1: (args.command, '')}, args.jobfile)
    commands = load_jobfile(args.jobfile)[0]

    backend = prepare_backend(args)

    # Update additional arguments
    if args.tasklist is None:
        args.tasklist = backend.generate_tasklist(commands)

    jobs = backend.get_job_list(args)

    if len(other_args) != 0 and not args.quiet:
        print("Passing these args to backend: {}".format(' '.join(other_args)))

    backend.launch(jobs, args, other_args)



import configparser
import os
from warnings import warn

from .constants import globalconfigfile, localconfigfile
from . import backends

defaults = {
    'DEFAULT': {
        'header': '',
        'footer': '',
    },
    'slurm': {},
    'gridengine': {},
}

def ensure_initialized():
    settings = configparser.ConfigParser()
    settings.read_dict(defaults)
    if not os.path.exists(localconfigfile):
        with open(localconfigfile, 'w') as config_file:
            settings.write(config_file)

def print_config(config):
    for section in config.sections():
        print('[{}]'.format(section))
        for key, value in dict(config[section]).items():
            quote = '"""'  if '\n' in value else ''
            print('{} = {}{}{}'.format(key, quote, value, quote))
        print()

def get_active_config():
    ensure_initialized()
    settings = configparser.ConfigParser()
    settings.read(globalconfigfile)
    settings.read(localconfigfile)
    return settings

def config(args):
    if args.read and args.write:
        raise RuntimeError('Cannot read and write at the same time.')
    if args.read and args.global_ and args.local:
        args.global_ = False
        args.local = False
    if args.write and args.global_ and args.local:
        raise RuntimeError('Cannot write to global and local config files simultaneously.')

    settings = configparser.ConfigParser()
    ensure_initialized()
    settings.read_dict(defaults)
    if args.global_ or not args.local:
        settings.read(globalconfigfile)
    if args.local or not args.global_:
        settings.read(localconfigfile)
    
    if args.read:
        print_config(settings)
    elif args.write:
        for (backend, key, value) in args.write:
            if backend not in backends.__all__:
                warn('Unable to set {}.{}={} (invalid backend: {}).'.format(backend, key, value, backend))
                continue
            # if key not in settings[backend].keys():
            #     warn('Unable to set {}.{}={} (invalid key: {}).'.format(backend, key, value, key))
            #     continue
            settings[backend][key] = value
        
        if args.global_:
            with open(globalconfigfile, 'w') as config_file:
                settings.write(config_file)
        else:
            with open(localconfigfile, 'w') as config_file:
                settings.write(config_file)


from collections import OrderedDict
import os
from warnings import warn

from .utils import load_jobfile, save_jobfile
from .constants import sep, wsep, flag_on, flag_off

def meta_launch(args):
    base_cmd = args.command

    if args.arg is not None:
        variables = OrderedDict({arglist[0]: arglist[1:] for arglist in args.arg})
    else:
        variables = OrderedDict()

    if args.pos_arg is not None:
        pos_variables = args.pos_arg
    else:
        pos_variables = []

    if args.flag is not None:
        flag_variables = args.flag
    else:
        flag_variables = []

    base_cmd_args = list(variables.keys())

    cmd_prefix_list = [base_cmd]

    if args.tag == '':
        raise ValueError("+tag cannot be an empty string")

    if args.tag is not None:
        cmd_suffix_list = ['']
        if args.tag_args is None:
            args.tag_args = base_cmd_args
        else:
            for tag_arg in args.tag_args:
                if tag_arg not in base_cmd_args:
                    warn(RuntimeWarning("{} is not a command arg: {}".format(tag_arg,
                        base_cmd_args)))


    # Positional arguments
    for value_list in pos_variables:
        cmd_prefix_list = [prefix + ' {}' for prefix in cmd_prefix_list]
        cmd_prefix_list = [prefix.format(v) for v in value_list for prefix in cmd_prefix_list]
        if args.tag is not None:
            value_slot = wsep + '{}'
            cmd_suffix_list = [
                suffix + value_slot for suffix in cmd_suffix_list
            ]
            cmd_suffix_list = [
                suffix.format(v) for v in value_list for suffix in cmd_suffix_list
            ]

    # Optional arguments
    for key, value_list in variables.items():
        cmd_prefix_list = [prefix + ' ' + key for prefix in cmd_prefix_list]
        if len(value_list) > 1:
            cmd_prefix_list = [prefix + ' {}' for prefix in cmd_prefix_list]
            cmd_prefix_list = [prefix.format(v) for v in value_list for prefix in cmd_prefix_list]
        if args.tag is not None:
            if key in args.tag_args:
                value_slot = sep + '{}' if len(value_list) > 1 else ''
                keyname = key.replace('_', '').replace('-', '')
                cmd_suffix_list = [
                    suffix + wsep + keyname + value_slot for suffix in cmd_suffix_list
                ]
                if len(value_list) > 1:
                    cmd_suffix_list = [
                        suffix.format(v) for v in value_list for suffix in cmd_suffix_list
                    ]
            else:
                cmd_suffix_list = [suffix for v in value_list for suffix in cmd_suffix_list]

    # Flag/Boolean arguments
    for flag in flag_variables:
        cmd_prefix_list = [prefix + ' {}' for prefix in cmd_prefix_list] + cmd_prefix_list
        cmd_prefix_list = [
            prefix.format(flag) if '{}' in prefix else prefix
            for prefix in cmd_prefix_list
        ]
        if args.tag is not None:
            cmd_suffix_list = [
                suffix + '{}' for suffix in cmd_suffix_list
            ]
            cmd_suffix_list = [
                suffix.format(wsep + s + flag.replace(flag_off, '').replace(flag_on, ''))
                for s in [flag_on, flag_off]
                for suffix in cmd_suffix_list
            ]

    jobfile_path = args.jobfile.format(jobname=args.jobname)
    os.makedirs(os.path.dirname(jobfile_path), exist_ok=True)

    if args.append:
        jobs, tags = load_jobfile(jobfile_path)
        start_jobid = max(jobs.keys()) + 1
    else:
        jobs = dict()
        start_jobid = 1

    if args.tag is not None:
        tag_list = [
            args.jobname + sep + '{}'.format(i) + suffix
            for (i, suffix) in enumerate(cmd_suffix_list, start_jobid)
        ]
        cmd_prefix_list = [
            ' '.join([prefix, args.tag, suffix])
            for (prefix, suffix) in zip(cmd_prefix_list, tag_list)
        ]
    else:
        tag_list = [""]*len(cmd_prefix_list)

    for i, (cmd,tag) in enumerate(zip(cmd_prefix_list,tag_list), start_jobid):
        if not args.quiet:
            print(cmd)
        jobs[i] = (cmd,tag)

    save_jobfile(jobs, jobfile_path, args.tag)



import subprocess
import sys

from .backends import prepare_backend
from .list import get_job_listings

def launch_cancel_proc(cmd, args):
    """Print the cancel command and launch a subprocess to execute it"""
    if not args.quiet:
        print(cmd)
    if not args.dry_run:
        try:
            subprocess.call(cmd, shell=True)
        except (subprocess.CalledProcessError, ValueError) as err:
            print(err)
            sys.exit()

def cancel(args):
    """Parse the jobs/tasks to cancel and send the appropriate commands to the cluster"""

    if args.tasklist is None:
        yes_or_no = input('Are you sure you want to cancel all tasks for this job? (y/[n])\n> ')
        if yes_or_no in ['y','yes','Y',"YES"]:
            pass
        else:
            if yes_or_no not in ['n','no','N',"NO",'']:
                print('Unable to process response "{}"'.format(yes_or_no))
            print('Job cancellation aborted.')
            sys.exit()

    job_list = get_job_listings(args)
    unique_job_ids = sorted(list(set([listing.job_id for listing in job_list])))

    tasklists = {job_id: [] for job_id in unique_job_ids}
    for listing in job_list:
        tasklists[listing.job_id].append(listing.task_id)

    cancellations = [(job_id, tasklists[job_id]) for job_id in unique_job_ids]

    backend = prepare_backend(args)
    cmds = backend.get_cancel_cmds(cancellations)
    
    for cmd in cmds:
        launch_cancel_proc(cmd, args)



from setuptools import setup, find_packages
setup(
    name='onager',
    version='0.1.0',
    description='Lightweight python library for launching experiments and tuning hyperparameters, either locally or on a cluster',
    long_description=open('README.md').read(),
    license=open('LICENSE').read(),
    author='Cameron Allen, Neev Parikh',
    author_email=('csal@brown.edu, neev_parikh@brown.edu'),
    packages=find_packages(include=['onager', 'onager.*']),
    scripts=['bin/onager'],
    url='http://pypi.python.org/pypi/onager/',
    install_requires=[
        "tabulate",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import os

# Basic constants
default_scripts_folder = os.path.join('.onager', 'scripts')
default_logs_folder = os.path.join('.onager', 'logs')
default_index = os.path.join('.onager', 'job_index.csv') # id,jobname,jobfile_path
defaultjobfile = os.path.join(default_scripts_folder, '{jobname}', 'jobs.json')
globalconfigfile = os.path.join(os.path.expanduser('~'), '.onagerconfig')
localconfigfile = os.path.join('.onager', 'config')

# Separators
sep = '_'
wsep = '__'
flag_on = '+'
flag_off = '-'



import csv
from itertools import count, groupby
import json

from .constants import default_index

def load_jobfile(jobfile_path):
    with open(jobfile_path, 'r') as file:
        job_records = json.load(file)
    # json stores all keys as strings, so we convert to ints
    jobs = {int(id_): record[0] for id_, record in job_records.items()}
    tags = {int(id_): record[1] for id_, record in job_records.items()}
    return jobs, tags

def save_jobfile(jobs, jobfile_path, tag=None):
    with open(jobfile_path, "w+") as jobfile:
        json.dump(jobs, jobfile)

def load_jobindex():
    with open(default_index, 'r', newline='') as job_index:
        csv_reader = csv.reader(job_index, delimiter=',', quotechar='|')
        index = {job_entry[0]: job_entry[1:] for job_entry in csv_reader}
    return index

def get_next_index_jobid():
    try:
        next_jobid = max(map(int,load_jobindex().keys())) + 1
    except IOError:
        next_jobid = 0
    return next_jobid

def update_jobindex(job_entries, append=True):
    mode = 'w+' if not append else 'a+'
    with open(default_index, mode, newline='') as job_index:
        csv_writer = csv.writer(job_index, delimiter=',', quotechar='|')
        csv_writer.writerows(job_entries)

def condense_ids(id_list):
    G = (list(x) for _, x in groupby(id_list, lambda x, c=count(): next(c) - x))
    return ",".join("-".join(map(str, (g[0], g[-1])[:len(g)])) for g in G)

def expand_ids(tasklist):
    return [i for r in _generate_id_ranges(tasklist) for i in r]

def _generate_id_ranges(tasklist):
    task_blocks = tasklist.split(',')
    for task_block in task_blocks:
        if ':' in task_block:
            task_block, step = task_block.split(':')
            step = int(step)
        else:
            step = 1
        if '-' in task_block:
            first, last = map(int, task_block.split('-'))
        else:
            first = int(task_block)
            last = first
        yield range(first, last + 1, step)

def insert_second_to_last(cmd, insert_str, sep=' '):
    cmd = cmd.split(sep)
    return sep.join(cmd[:-1]) + sep + insert_str + sep + cmd[-1]
